use proc_macro::TokenStream;
use proc_macro2::TokenStream as TokenStream2;
use quote::quote;

use update_macro_base::*;

pub fn resource_only_events_and_systems(update: &Update) -> TokenStream2 {
    let resource_tokens = update.resourece_tokens();

    let resource_types = &resource_tokens.resource_types;
    let resource_type_impls = &resource_tokens.resource_type_impls;
    let resources = &resource_tokens.resources;
    let resource_requirement = &resource_tokens.resource_requirement;
    let resource_store = &resource_tokens.resource_store;
    let resource_idents = &resource_tokens.resource_idents;

    quote! {
        impl<P, Func, #resource_type_impls> EventReader<P, (), #resource_types, Func, ()> for Events
        where
            Func: Fn(&P, &mut Commands, #resources ) -> Result<()> + Send + Sync + 'static,
            #resource_requirement
            P: std::any::Any + Send + Sync,
        {
            fn add_reader(&mut self, func: Func) {
                match self.events.get_mut(&TypeId::of::<P>()) {
                    Some((_, listener)) => listener.push(std::sync::Arc::new(move |world, payload| {
                        let typed_payload: &P = payload.downcast_ref().unwrap();
                        let mut commands = Commands::new(world.now());

                        #resource_store
                        func(typed_payload, &mut commands, #( #resource_idents, )*)?;

                        commands.apply_deferred(world)?;

                        Ok(())
                    })),
                    None => panic!("register event type first!"),
                }
            }
        }

        impl<Func, #resource_type_impls> AddSystem<#resource_types, Func, ()> for WorldBuilder
        where
            Func: Fn(&mut Commands, #resources ) -> Result<bool> + Send + Sync + 'static,
            #resource_requirement
        {
            fn add_system(&mut self, priority: u32, func: Func) {
                self.systems.push((priority, Box::new(move |world: &mut World, commands: &mut Commands| {
                    #resource_store
                    func(commands, #( #resource_idents, )*)
                })));
                self.systems.sort_by_key(|(priority, _)| *priority);
            }
        }

        impl<Func, #resource_type_impls> AddSystem<#resource_types, Func, World> for WorldBuilder
        where
            Func: Fn(&World, &mut Commands, #resources ) -> Result<bool> + Send + Sync + 'static,
            #resource_requirement
        {
            fn add_system(&mut self, priority: u32, func: Func) {
                self.systems.push((priority, Box::new(move |world: &mut World, commands: &mut Commands| {
                    let w = unsafe { utilities::unsafe_life_time::remove_life_time(world) };
                    #resource_store
                    func(w, commands, #( #resource_idents, )*)
                })));
                self.systems.sort_by_key(|(priority, _)| *priority);
            }
        }
    }
}

fn events_and_systems(update: &Update) -> TokenStream2 {
    let query_tokens = update.query_tokens();

    let filter_types = &query_tokens.filter_types;
    let filter_type_impls = &query_tokens.filter_type_impls;
    let query_types = &query_tokens.query_types;
    let query_type_impls = &query_tokens.query_type_impls;
    let queries = query_tokens.queries;
    let filter_requirements = &query_tokens.filter_requirements;
    let component_requirements = &query_tokens.component_requirements;
    let query_count = query_tokens.query_count;
    let verify_dedup = &query_tokens.verify_dedup;
    let check_entities = &query_tokens.check_entities;
    let component_stores = &query_tokens.component_stores;
    let query_structs = &query_tokens.query_structs;
    let query_idents = &query_tokens.query_idents;

    let resource_tokens = update.resourece_tokens();

    let resource_types = &resource_tokens.resource_types;
    let resource_type_impls = &resource_tokens.resource_type_impls;
    let resources = &resource_tokens.resources;
    let resource_requirement = &resource_tokens.resource_requirement;
    let resource_store = &resource_tokens.resource_store;
    let resource_idents = &resource_tokens.resource_idents;

    quote! {
        impl<Func, #filter_type_impls #query_type_impls #resource_type_impls> CreateArchetype<#query_types, #resource_types, Func, #filter_types> for Archetype
        where
            Func: Fn(&mut Commands, #( #queries, )* #resources ) -> Result<()> + Send + Sync + Clone + 'static,
            #filter_requirements
            #component_requirements
            #resource_requirement
        {
            fn create(f: Func) -> Self {
                #( #verify_dedup )*

                Self {
                    queries: #query_count,

                    check_entity: Box::new(move |entity| {
                        let mut comply_queries = Vec::new();

                        #( #check_entities )*

                        if comply_queries.is_empty() {
                            None
                        } else {
                            Some(comply_queries)
                        }
                    }),

                    create_callback: Box::new(move |entities| {
                        #( #component_stores )*

                        let f = f.clone();

                        Ok(Box::new(move |commands, world| {
                            #( #query_structs )*
                            #resource_store

                            f(commands, #( #query_idents, )* #( #resource_idents, )* )
                        }))
                    }),

                    entities: IndexMap::new(),
                }
            }
        }

        impl<Func, #filter_type_impls #query_type_impls #resource_type_impls> AddUpdates<#query_types, #resource_types, Func, #filter_types> for Updates
        where
            Func: Fn(&mut Commands, #( #queries, )* #resources ) -> Result<()> + Send + Sync + Clone + 'static,
            #filter_requirements
            #component_requirements
            #resource_requirement
        {
            fn add_update(
                &mut self,
                priority: u32,
                func: Func,
            ) -> Result<()> {
                self.add(priority, Archetype::create(func))
            }
        }

        impl<Func, #filter_type_impls #query_type_impls #resource_type_impls> AddUpdates<#query_types, #resource_types, Func, #filter_types> for WorldBuilder
        where
            Func: Fn(&mut Commands, #( #queries, )* #resources ) -> Result<()> + Send + Sync + Clone + 'static,
            #filter_requirements
            #component_requirements
            #resource_requirement
        {
            fn add_update(
                &mut self,
                priority: u32,
                func: Func,
            ) -> Result<()> {
                self.updates.add_update(priority, func)
            }
        }

        impl<P, Func, #filter_type_impls #query_type_impls #resource_type_impls> EventReader<P, #query_types, #resource_types, Func, #filter_types> for Events
        where
            Func: Fn(&P, &mut Commands, #( #queries, )* #resources ) -> Result<()> + Send + Sync + 'static,
            #filter_requirements
            #component_requirements
            #resource_requirement
            P: std::any::Any + Send + Sync,
        {
            fn add_reader(&mut self, func: Func) {
                #( #verify_dedup )*

                match self.events.get_mut(&TypeId::of::<P>()) {
                    Some((_, listener)) => listener.push(std::sync::Arc::new(move |world, payload| {
                        let typed_payload: &P = payload.downcast_ref().unwrap();
                        let mut commands = Commands::new(world.now());

                        #resource_store

                        let mut complying_entities: Vec<Vec<&EntityObject>> = vec![Vec::new(); #query_count];

                        for entity in world.entities.values() {
                            let mut comply_queries = Vec::new();

                            #( #check_entities )*

                            if !comply_queries.is_empty() {
                                comply_queries
                                    .into_iter()
                                    .for_each(|q| complying_entities[q].push(entity));
                            }
                        }

                        // check if one of the queries could not be fullfilled
                        for queryable_entities in complying_entities.iter() {
                            if queryable_entities.is_empty() {
                                return Ok(());
                            }
                        }

                        for entities in complying_entities.into_iter().multi_cartesian_product() {
                            #( #component_stores )*
                            #( #query_structs )*

                            func(typed_payload, &mut commands, #( #query_idents, )* #( #resource_idents, )*)?;
                        }

                        commands.apply_deferred(world)?;

                        Ok(())
                    })),
                    None => panic!("register event type first!"),
                }
            }
        }
    }
}

pub fn update(max_components: usize, max_resources: usize) -> TokenStream {
    let mut queries: Vec<Vec<usize>> = Vec::new();

    for c in Update::MIN_COMPONENTS..=max_components {
        queries.push(vec![c]);
    }

    for x in Update::MIN_COMPONENTS..=max_components {
        for y in Update::MIN_COMPONENTS..=max_components {
            queries.push(vec![x, y]);
        }
    }

    for x in Update::MIN_COMPONENTS..=max_components {
        for y in Update::MIN_COMPONENTS..=max_components {
            for z in Update::MIN_COMPONENTS..=max_components {
                queries.push(vec![x, y, z]);
            }
        }
    }

    let mut resources = Vec::new();

    for resource in Update::MIN_RESOURCE..=max_resources {
        resources.push(resource);
    }

    let mut updates = Vec::new();

    for query in queries.iter() {
        for resource_count in resources.iter() {
            updates.push(events_and_systems(&Update::new(
                query.iter().cloned(),
                *resource_count,
            )));
        }
    }

    let mut events = Vec::new();

    for resource_count in resources.iter() {
        let q =
            resource_only_events_and_systems(&Update::new(vec![1].into_iter(), *resource_count));

        events.push(q);
    }

    // updates.push(Update::new(vec![1].into_iter(), 0));

    // let q = quote! {
    //     #( #updates )*
    // };

    // panic!("{q}");

    TokenStream::from(quote! {
        #( #updates )*
        #( #events )*
    })
}
